# -*- coding: utf-8 -*-
"""
James Elgy - 2019
Overarching data class to deal with the loading and manipulation of GBSAR data.
"""
import numpy as np
import tkinter as tk
from tkinter.filedialog import askopenfilename as aof
from GBSAR.Windows import Window


class Data:

    def __init__(self, FileName='OpenDialogue'):
        self.FileName = 'OpenDialogue'

    def LoadFile(self):
        """
        James Elgy - 2019

        GBSAR Ascii file reader. Originally written in MATLAB.

        This is a function to open and read the standard format of GBSAR .scn
        files.
        The function requires that the full path name of the file in question
        is entered as the function argument.
        It cycles through the header names and records the values as numbers
        in a dictionary. E.g. Data[Positions] = 101.
        It records the complex amplitude at each frequency and aperture
        position into the matrix 'RawData'.

        The function requires that the input be a string containing the
        file name. E.g. 'C:\\Users\\.....\\EMScan171103_307_0001
        """
        # Getting FileName

        if self.FileName == 'OpenDialogue':
            root = tk.Tk()
            root.withdraw()
            self.FileName = aof(title="Select file", filetypes=(("Scan Files", "*.scn"), ("all files", "*.*")))

        File = open(self.FileName, 'r')
        Line = File.readline()

        # Finding Rail Settings
        if self.FileName[-4:] == '.cal':
            self.Positions = 1
            self.FileType = 'cal'
        else:
            while Line[0:9] != 'Positions':
                Line = File.readline()
            self.Positions = int(Line[12:-1])
            self.FileType = 'scn'

        # Here, I divide by 1000 since I want to convert from mm to m.
        while Line[0:18] != 'HorizontalAperture':
            Line = File.readline()
        self.HorizontalAperture = float(Line[21:-1])/1000

        while Line[0:23] != 'HorizontalStartPosition':
            Line = File.readline()
        self.HorizontalStartPosition = float(Line[26:-1])/1000

        while Line[0:19] != 'HorizontalIncrement':
            Line = File.readline()
        self.HorizontalIncrement = float(Line[22:-1])/1000

        while Line[0:16] != 'VerticalAperture':
            Line = File.readline()
        self.VerticalAperture = float(Line[19:-1])/1000

        while Line[0:21] != 'VerticalStartPosition':
            Line = File.readline()
        self.VerticalStartPosition = float(Line[24:-1])/1000

        while Line[0:17] != 'VerticalIncrement':
            Line = File.readline()
        self.VerticalIncrement = float(Line[20:-1])/1000

        # Finding VNA Settings
        while Line[0:14] != 'StartFrequency':
            Line = File.readline()
        self.StartFrequency = float(Line[17:-1])

        while Line[0:13] != 'StopFrequency':
            Line = File.readline()
        self.StopFrequency = float(Line[16:-1])

        while Line[0:14] != 'NumberOfPoints':
            Line = File.readline()
        self.FrequencyPoints = int(Line[17:-1])

        # Loading in Data
        AntennaArray = np.empty([2, self.Positions], dtype=float)
        RawData = np.empty([self.FrequencyPoints, self.Positions],
                           dtype=complex)

        for Index in range(self.Positions):
            while Line[0:8] != '[Reading':
                Line = File.readline()
            Line = File.readline()
            if self.FileType == 'scn':
                AntennaArray[:, Index] = np.array([float(Line.split(' ')[0]),
                                                   float(Line.split(' ')[1])])
                AntennaArray[:,Index] /= 1000
                File.readline()

            File.readline()  # Need to clean this up.
            File.readline()
            File.readline()
            File.readline()
            for Freq in range(self.FrequencyPoints):
                Line = File.readline()
                DataLine = list(filter(None, Line.split(' ')))
                RawData[Freq, Index] = float(DataLine[0]) + \
                    (float(DataLine[1])*1j)

        self.AntennaArray = AntennaArray
        self.RawData = RawData

        self.FrequencyArray = np.linspace(self.StartFrequency,
                                          self.StopFrequency,
                                          self.FrequencyPoints)

        File.close()

    def FormDataCube(self):
        """
        James Elgy - 2019
        This function reforms the 2d array generated by LoadFile into a 3d cube
        of size NFreqsXNThetaXNPhi. The motivation behind this function is that
        it makes apply 3D windows to the data significantly easier.
        """
        if np.all(np.diff(self.AntennaArray[1, :]) == 0):
            NHoriz = int(self.HorizontalAperture/self.HorizontalIncrement + 1)
            NVert = 1
        else:
            NHoriz = int(self.HorizontalAperture/self.HorizontalIncrement + 1)
            NVert = int(self.VerticalAperture/self.VerticalIncrement + 1)
            DataCube = np.reshape(self.RawData, (self.FrequencyPoints,
                                                 NHoriz, NVert))
            self.RawData = DataCube

    def UnFormDataCube(self):
        self.RawData = np.reshape(self.RawData,
                                  (self.FrequencyPoints, self.Positions))

    def ApplyWindow(self, WinType, Dims=1):
        """
        James Elgy - 2019
        This function applys a window (defined by WinType as a string)
        to the raw data. For example a Hamming window.
        The function enables for the application of a specific window
        (see Windows.py for details) in either 1, 2, or 3 dimensions.
        The number of dimensions is Dims
        """

        if Dims == 1:
            Size = (self.RawData.shape[0], 1, 1)
        elif Dims == 2:
            Size = (self.RawData.shape[0], self.RawData.shape[1], 1)
        elif Dims == 3:
            if len(self.RawData.shape) < 3:
                self.FormDataCube()

            Size = (self.RawData.shape[0],
                    self.RawData.shape[1],
                    self.RawData.shape[2])

        W = Window(WinType, Size)
        W.GetWindow()
        Win = np.resize(np.squeeze(W.Window), self.RawData.shape)
        self.RawData = self.RawData * Win
        self.Window = W

    def SubtractCL(self, CableLength):
        """
        James Elgy - 2019
        This function takes an input value of Controls['CableLength'] and
        shifts the range profiles down range by that amount.
        """

        if len(self.RawData.shape) == 3:
            self.UnFormDataCube()
            Flag = 1
        else:
            Flag = 0

        c = 299792458/1.0003
        Freq = self.FrequencyArray
        Freq.shape = (self.FrequencyPoints, 1)
        Const = 1j*4*np.pi*Freq
        PhaseRamp = np.exp(Const*(CableLength*np.ones([1, self.Positions]))/c)
        self.RawData = self.RawData * PhaseRamp

        if Flag == 1:
            self.FormDataCube()

    def SubtractMean(self):
        """
        James Elgy - 2019
        Function to subtract the mean value -in azimuth- from each pulse
        of the SAR data.
        """
        M = np.mean(self.RawData, 1)
        M.shape = (self.FrequencyPoints, 1)
        self.RawData = self.RawData - M

    def GetGeometry(self, Params):
        """
        James Elgy - 2019
        This function is designed to convert between the emscan coordinates to
        a ground truth coordinate system. The origin of both systems is kept
        the same.
        """

        Ant1 = np.zeros([3, self.Positions])
        Ant1[0] = self.AntennaArray[0]
        Ant1[2] = Params.RadarHeight - self.AntennaArray[1]

        Theta = np.deg2rad(Params.RotationAngle)
        R = np.array([[np.cos(Theta), -np.sin(Theta), 0],
                      [np.sin(Theta), np.cos(Theta), 0],
                      [0, 0, 1]])

        Grabber = geometrygrabber()
        Ant2_temp = np.asarray(Grabber.getGeometry(Params.GeometryIndex))

        if Params.GeometryIndex == 0:
            Ant2_temp = np.matmul(R, Ant2_temp)
            Ant1 = np.matmul(R, Ant1)
            Ant2_temp.shape = [3, 1]
            Ant2 = Ant1 + Ant2_temp
        else:
            Ant2_temp.shape = [3, 1]
            Ant2 = Ant2_temp * np.ones([1, Ant1.shape[1]])
            Ant1 = np.matmul(R, Ant1)

        return (Ant1, Ant2)

    def CorrectSC(self, Params):
        """
        James Elgy - 2019
        Function to calculate the range from the antennas to the scene centre.
        This populates Data['ScRange']. THis range is then used to shift the
        Range profiles.
        """

        c = 299792458/1.0003
        Sc = Params.SC
        Freq = self.FrequencyArray
        Freq.shape = [self.FrequencyPoints,  1]

        Sc.shape = [3, 1]
        Ant1, Ant2 = self.GetGeometry(Params)
        ScRange = np.zeros([self.Positions])

        Tsc = Sc - Ant1
        Rsc = Sc - Ant2
        for pos in range(0, self.Positions):
            Rt = np.sqrt(Tsc[:, pos].dot(Tsc[:, pos]))
            Rr = np.sqrt(Rsc[:, pos].dot(Rsc[:, pos]))
            ScRange[pos] = np.mean([Rt, Rr])

        ScRange.shape = [1, self.Positions]
        PhaseRamp = np.exp(1j * 4 * np.pi * Freq * ScRange / c)

        self.RawData = self.RawData * PhaseRamp
        self.ScRange = ScRange


class geometrygrabber(object):
    """
    Class to define the receiving antenna geometry.
    I chose to put it as a class to give a greater degree of flexability
    over the positioning.
    """

    def getGeometry(self, argument):
        caseName = 'Geo_' + str(argument)
        case = getattr(self, caseName, lambda: "invalid")
        return case()

    def Geo_0(self):
        Ant2 = [0, 0, 0]
        return Ant2

    def Geo_1(self):
        Ant2 = [-1.001, 0.765, 0.229]
        return Ant2

    def Geo_2(self):
        Ant2 = [-1.335, 0.458, 0.234]
        return Ant2

    def Geo_3(self):
        Ant2 = [-1.4459, 0.6053, 0.234]
        return Ant2

    def Geo_4(self):
        Ant2 = [-0.988, 0.816, 0.229]
        return Ant2

    def Geo_5(self):
        Ant2 = [-1.3373, 0.4717, 0.234]
        return Ant2

    def Geo_6(self):
        Ant2 = [-1.4822, 0.6235, 0.234]
        return Ant2

    def Geo_7(self):
        Ant2 = [-1.0196, 0.8179, 0.2308]
        return Ant2
