"""
James Elgy - 2019
Module for the formation and processing of monostatic and bistatic SAR
imagery on the graphics card.
This version has been slightly edited in order for it to be called from
and integrate with matlab 2019.

The MSAR2 module has some slight variations from the original MSAR module.
Namely that the for loop used in the backprojection has been moved onto
the gpu. This is to enable as much of the range profiles to be dumped onto
the gpu in one go as possible.
This change improved the execution time by approx 100% for the few tests
I've done.

As it stands, the entire image has to be dumped onto the GPU. Therefore
very large images need to be segmented and formed seperately.

Note:
The environment variable 'CUDA_HOME' is currently hardcoded. Make sure it
matches with the appropriate directory.
"""

import cmath
import math
import os

import numpy as np
from numba import cuda

def BP_GPU(xm, ym, zm, RPvec, Ant1, Ant2, Freq, RangeSc, RangeProfile_Real, RangeProfile_Imag):

    os.environ['CUDA_HOME'] = 'C:\\Program Files\\NVIDIA GPU Computing Toolkit\\CUDA\\v10.0'
    

	# This block of code has been added to convert between Matlab datatypes and NumPy.
    xm = np.ascontiguousarray(np.array(xm))
    ym = np.ascontiguousarray(np.array(ym))
    zm = np.ascontiguousarray(np.array(zm))
    Ant1 = np.array(Ant1)
    Ant2 = np.array(Ant2)
    Freq = np.array(Freq)
    RangeSc = np.array(RangeSc)
    RPR = np.array(RangeProfile_Real)
    RPC = np.array(RangeProfile_Imag)
    RPvec = np.array(RPvec)
    RangeProfile = RPR + 1j*RPC

	#Reshaping the imaging into a 1xN array reduces helps with indexing on the GPU
    OrigSize = xm.shape
    xm.shape = [np.prod(xm.shape), 1]
    ym.shape = [np.prod(ym.shape), 1]
    zm.shape = [np.prod(zm.shape), 1]

    ComplexImage = np.zeros(xm.shape, dtype=np.complex64)
    Pixel = np.ascontiguousarray(np.concatenate((xm, ym, zm), axis=1))
    Image = np.ascontiguousarray(np.zeros(xm.shape), dtype=np.complex64)

    TPB, BPG = Initalise(np.prod(xm.shape))
    # Here, I'm using stream to set up an asynchronous execution.
    Stream = cuda.stream()
    Cont = cuda.current_context()
    

    Pixel = cuda.to_device(Pixel, stream=Stream)
    Image = cuda.to_device(Image, stream=Stream)
    Freq = np.median(Freq)

    R0 = RPvec[0]
    Rdiff = np.diff(RPvec[0:2])

    # I've added in some functionality to split the range profiles up and dump
    # as much as possible onto the GPU. I've hardcoded a proportion cut off for
    # the free memory. I.e if the memory needed to dump the entire thing is
    # greater than 0.5* the amount of free memory, it will split the range
    # profiles up and iterate though.

    # Chunking Range Profiles
    Mem = Cont.get_memory_info()
    RPChunkSize = RangeProfile.nbytes/(0.5*Mem[0])

    if RPChunkSize <= 1:
        RP = cuda.to_device(np.ascontiguousarray(np.conj(RangeProfile)), stream=Stream)
        RangeSc_Pos = cuda.to_device(np.ascontiguousarray(RangeSc), stream=Stream)
        A1 = Ant1
        A2 = Ant2
        A1 = cuda.to_device(np.ascontiguousarray(A1), stream=Stream)
        A2 = cuda.to_device(np.ascontiguousarray(A2), stream=Stream)

        interpKernel[BPG, TPB, Stream](Pixel, A1, A2, RangeSc_Pos, Freq, R0,
                                         Rdiff, RP, Image)
        ComplexImage = Image.copy_to_host()

    else:
        # This else statement is included so that datasets larger than the GPU
        # can be used.
        NChunks = np.array_split(RangeProfile,
                                 int(math.ceil(RPChunkSize)), axis=1)

        for Index in range(0, int(math.ceil(RPChunkSize))-1):
            if Index == 0:
                CountOld = 0
                Count = NChunks[Index].shape[1]
            else:
                CountOld = Count
                Count = Count + NChunks[Index].shape[1]

            print(Count)
            RP_Contig = np.ascontiguousarray(np.conj(NChunks[Index]))
            RP = cuda.to_device(RP_Contig)
            RangeSc_Contig = np.ascontiguousarray(RangeSc[CountOld:Count])
            RangeSc_Pos = cuda.to_device(RangeSc_Contig)
            A1 = Ant1[:, CountOld:Count]
            A2 = Ant2[:, CountOld:Count]
            A1 = cuda.to_device(np.ascontiguousarray(A1))
            A2 = cuda.to_device(np.ascontiguousarray(A2))
            Stream.synchronize()
            interpKernel[BPG, TPB](Pixel, A1, A2, RangeSc_Pos, Freq,
                                     R0, Rdiff, RP, Image)

    Stream.synchronize()
    ComplexImage = np.reshape(ComplexImage, OrigSize)
    ComplexImage = np.squeeze(ComplexImage)

    ComplexImageReal = np.ascontiguousarray(np.real(ComplexImage))
    ComplexImageImag = np.ascontiguousarray(np.imag(ComplexImage))

    cuda.close() # Closing the current context to avoid errors.

    return (ComplexImageReal, ComplexImageImag)


def Initalise(imsize):
    # Set the number of threads in a block
	# Supposedly, changing this will change performance however I've 
	# noticed no difference.
    threadsperblock = 32
    # Calculate the number of thread blocks in the grid
    blockspergrid = (imsize + (threadsperblock - 1)) // threadsperblock

    return (threadsperblock, blockspergrid)


@cuda.jit()
def interpKernel(Pixel, A1, A2, RangeSc, Freq, R0, Rdiff, RP, Image):
    # Second version of the backprojection kernel. This version has the for
    # loop moved inside the kernel. This allows for more data to be dumped onto
    # the gpu in one go, thus reducing transfer times.

    Ele = cuda.grid(1)
    PhaseConst = -1j * 4 * Freq * cmath.pi / 3e8
    # if pos < Image.size:  # Check array boundaries
    for Pulse in range(0, RP.shape[1]):
        # Doing backprojection for each pixel in image.
        dr = (math.sqrt((A1[0, Pulse] - Pixel[Ele, 0])**2 +
                        (A1[1, Pulse] - Pixel[Ele, 1])**2 +
                        (A1[2, Pulse] - Pixel[Ele, 2])**2) +
              math.sqrt((A2[0, Pulse] - Pixel[Ele, 0])**2 +
                        (A2[1, Pulse] - Pixel[Ele, 1])**2 +
                        (A2[2, Pulse] - Pixel[Ele, 2])**2))/2 - RangeSc[Pulse]

        Correction = cmath.exp(PhaseConst * dr)

        # Linear interp section. Doing it like this avoids the search
        # algorithms used in numpy.
        Ind = int((dr - R0)//Rdiff[0])
        Rv_min = R0 + Ind*Rdiff[0]
        Rv_max = R0 + (Ind+1)*Rdiff[0]
        Grad = (RP[Ind+1, Pulse] - RP[Ind, Pulse]) / (Rv_max - Rv_min)
        Interpval = RP[Ind, Pulse] + (Grad*(dr - Rv_min))

        # Updating image.
        cuda.syncthreads()
        Image[Ele, 0] = Image[Ele, 0] + (Interpval * Correction)
