function ComplexImage = PythonBP(xm, ym, zm, RPvec, Ant1, Ant2, Freq, RangeSc, RangeProfile_Real, RangeProfile_Imag)
% James Elgy - 2019
% 
% Function to perform direct backprojection on the GPU.
% This function relies on a python module MSAR to do the actual computation,
% however the MatlabHook function can still be called within Matlab.
% 
% Both numpy and numba must be installed along with a python executable. The
% path to this python executable can be defined using PYVERSION
% For more info - <a href="matlab:
% web('https://uk.mathworks.com/products/matlab/matlab-and-python.html')">Python in Matlab</a>
% 
% INPUT:
% - xm, ym, zm - the meshgrids defining the imaging space.
% - RPvec - The range profile vector.
% - Ant1, Ant2 - The 3 by N antenna coordinates.
% - Freq - The array of frequencies used in the collection.
% - RangeSc - The bistatic range from antennas to scene centre.
% - RangeProfile_Real - The real component of the complex range profile data.
% - RangeProfile_Imag - The imaginary component of the range profile data.
% 
% OUTPUT:
% - ComplexImage - The complex SAR image.


% Clearing gpu memory - Make more space.
G = gpuDevice(1);
reset(G);

% Adds current working directory to python path.
if count(py.sys.path,'') == 0
    insert(py.sys.path,int32(0),'');
end

% Importing Module
Mod = py.importlib.import_module('MSAR');
py.importlib.reload(Mod);

% Executing Backprojection. The output is a 2x1 tuple which is then
% converted to double.
C = Mod.BP_GPU(xm, ym, zm, RPvec, Ant1, Ant2, Freq, RangeSc, RangeProfile_Real, RangeProfile_Imag);

CR = double(C{1});
CI = double(C{2});
ComplexImage = CR - 1i*CI;

ComplexImage = ComplexImage.';

end
