# -*- coding: utf-8 -*-
"""
James Elgy - 2019
This module, as part of the GBSAR package, deals with the formation and manipulation
of range profiles.
As a class, RangeProfiles has several methods associated with it, the most important
being 'FormRangeProfiles'. 
"""
import numpy as np
from matplotlib import pyplot as plt


class RangeProfiles(object):

    def __init__(self, Data, Params):
        self.Params = Params
        self.Data = Data

    def FormRangeProfiles(self):

        """
        James Elgy 2019

        Function for calculation and plotting the range profile (B-scan) data.
        This scrip uses the phase history of the scene and zero pads it upto a
        length of Nfft. The range vector is calculated by calculating the
        maximum unambiguous range. The padded phase history is then converted
        into the RP data by means of a ifft.

        The script requires that:
         - RawData be populated with the correct phase history data.
         This is a 2D matrix of the signal amplitudes at all the scan positions
         and frequencies.
         - FrequencyArray be populated by the frequencies used in calculating
         the phase history.
         - UpsampleFactor a scalar number corresponding to the final size of the
         padded dataset.
         """

        c = 299792458/1.0003  # Speed of light in air.
        Bandwidth = max(self.Data.FrequencyArray) \
            - min(self.Data.FrequencyArray)

        DeltaF = Bandwidth / len(self.Data.FrequencyArray)
        Wr = c/(2*DeltaF)
        Nfft = len(self.Data.FrequencyArray) * self.Params.UpsampleFactor

        RangeVector = np.linspace(-Nfft/2, Nfft/2, num=Nfft) * (Wr/Nfft)

        PaddedData = np.zeros([Nfft, self.Data.Positions], dtype=complex)
        PaddedData[0: len(self.Data.FrequencyArray), :] = self.Data.RawData

        # Padded phase history is rolled over to centre around the origin.
        PaddedData = np.roll(PaddedData,
                             int(np.floor(-len(self.Data.FrequencyArray)/2)),
                             axis=0)

        RangeProfile = np.fft.fftshift(np.fft.ifft(PaddedData, axis=0), axes=0)

        self.RangeProfile = RangeProfile
        self.RangeVector = RangeVector

    def PlotRangeProfiles(self):
        
        """
        James Elgy - 2019
        Function to plot the complex range profile data in a dB scale.
        If the range profiles have not already been formed, then this function
        will do so automatically.
        
        NOTE:
        This function currently uses the matplotlib latex interpreter for its 
        axes labels.

        """
        
        
        try:
            self.RangeProfile
        except AttributeError:
            self.FormRangeProfiles()

        plt.rc('text', usetex=True)

        fig = plt.figure()
        plt.inferno()
        ax = fig.add_subplot(111)
        extent = (1, self.RangeProfile.shape[1],
                  min(self.RangeVector),
                  max(self.RangeVector))

        RPLog = 20*np.log10(np.abs(self.RangeProfile))

        ax.imshow(RPLog, vmin=np.max(RPLog)-20, vmax=np.max(RPLog),
                  origin='lower', extent=extent)

        ax.set_aspect('auto', 'box')
        ax.set_xlabel('Antenna Position Ind')
        ax.set_ylabel('Range (m)')
        ax.set_title('Range Profile')

    def RangeGate(self, Min, Max):

        """
        James Elgy - 2019
        This function crops the range profile data such that only
        values between a range of 'Min' and 'Max' are included. All
        else is removed.
        """
        
        try:
            self.RangeVector
        except AttributeError:
            print('Range profiles must be formed first')
            return

        Max = float(Max)
        Min = float(Min)
        GateMax = np.greater(self.RangeVector, Min)
        GateMin = np.less(self.RangeVector, Max)
        Gate = np.asmatrix(GateMax * GateMin, dtype=int)
#        Gate = np.transpose(Gate)
        Elements = np.nonzero(Gate)[1]
        self.RangeProfile = self.RangeProfile[Elements[0]:Elements[-1],:]
        self.RangeVector = self.RangeVector[Elements[0]:Elements[-1]]
        return Elements
