"""
James Elgy - 2019
Module for the formation and processing of monostatic and bistatic SAR
imagery.
The SAR class is designed to act as a repository for the signal processing
required to go from range profile data towards 2D and 3D SAR images.

"""

import cmath
import math

import numpy as np
from numba import jit, float64, complex128, prange, cuda
from matplotlib import pyplot as plt
from mayavi import mlab
import skimage
from scipy.constants import c


class SAR(object):

    def __init__(self, Data, Params, RangeProfiles):
        self.Params = Params
        self.Data = Data
        self.RangeProfiles = RangeProfiles

    def FormImagingGrid(self):
        """
        James Elgy - 2019
        Function to produce the x y and z meshgrids needed for backprojection.
        """

        Sc = self.Params.SC
        NumPixelsX = np.round(self.Params.Lcr / self.Params.CRDensity)
        NumPixelsY = np.round(self.Params.Lr / self.Params.RDensity)

        PixelsX = np.linspace(-self.Params.Lcr/2,
                              self.Params.Lcr/2,
                              NumPixelsX) + Sc[0]
        PixelsY = np.linspace(-self.Params.Lr/2,
                              self.Params.Lr/2,
                              NumPixelsY) + Sc[1]

        if self.Params.Plot3D:
            NumPixelsZ = np.round(self.Params.Lel / self.Params.ZDensity)
            PixelsZ = np.linspace(-self.Params.Lel/2,
                                  self.Params.Lel/2,
                                  NumPixelsZ) + Sc[2]
        else:
            NumPixelsZ = 1
            PixelsZ = self.Params.SC[2]

        xm, ym, zm = np.meshgrid(PixelsX, PixelsY, PixelsZ)

        return (xm, ym, zm)

    def BackProjection_GPU(self):
        """
        James Elgy - 2019
        Time domain backprojection function on GPU. The function requires that
        the raw data be corrected for the scene centre.
        The output of the function is self.ComplexImage.
        """

        try:
            self.RangeProfiles.RangeProfile
        except AttributeError:
            self.RangeProfiles.FormRangeProfiles()

        xm, ym, zm = self.FormImagingGrid()

        # Defining antenna geometry and frequency
        A1, A2 = self.Data.GetGeometry(self.Params)
        Freq = self.Data.FrequencyArray

        # Runnning Interpolator#

        self.ComplexImage = _BP_GPU_2(xm, ym, zm, self.RangeProfiles.RangeVector,
                                    A1, A2, Freq, self.Data.ScRange[0, :],
                                    self.RangeProfiles.RangeProfile)

    def BackProjection_CPU(self):
        """
        James Elgy - 2019
        Parallel CPU equivalent of BackProjection_GPU.
        """

        try:
            self.RangeProfiles.RangeProfile
        except AttributeError:
            self.RangeProfiles.FormRangeProfiles()

        xm, ym, zm = self.FormImagingGrid()

        # Defining antenna geometry and frequency
        A1, A2 = self.Data.GetGeometry(self.Params)
        Freq = self.Data.FrequencyArray

        # Runnning Interpolator
        self.ComplexImage = _BP_CPU(xm, ym, zm, self.RangeProfiles.RangeVector,
                                    A1, A2, Freq, self.Data.ScRange[0, :],
                                    self.RangeProfiles.RangeProfile)

    def PlotSAR(self):
        """
        James Elgy - 2019
        Function to plot 2d or 3d complex sar imagry on a dB scale.
        """
        try:
            self.ComplexImage
        except AttributeError:
            self.BackProjection_CPU()

        ComplexImage = np.squeeze(self.ComplexImage)
        xm, ym, zm = self.FormImagingGrid()

        if len(ComplexImage.shape) < 3:
            plt.rc('text', usetex=True)

            fig = plt.figure()
            plt.inferno()
            ax = fig.add_subplot(111)
            extent = (np.min(xm), np.max(xm), np.min(ym), np.max(ym))
            ImLog = 20*np.log10(np.abs(ComplexImage))

            ax.imshow(ImLog, vmin=np.max(ImLog)-self.Params.DynamicRange,
                      vmax=np.max(ImLog), origin='lower', extent=extent)
            ax.set_aspect('auto', 'box')
            ax.set_xlabel('Cross Range (m)')
            ax.set_ylabel('Range (m)')
            ax.set_title('SAR Image')
        else:
            ImLog = 20*np.log10(np.abs(self.ComplexImage))
            VMax = np.max(ImLog)
            VMin = VMax - self.Params.DynamicRange
            mlab.figure(bgcolor=(1, 1, 1), fgcolor=(0, 0, 0))
            mlab.pipeline.volume(mlab.pipeline.scalar_field(ImLog),
                                 vmin=VMin, vmax=VMax)
            mlab.outline()
            Ranges = [np.min(xm), np.max(xm), np.min(ym), np.max(ym),
                      np.min(zm), np.max(zm)]
            Ax = mlab.axes(ranges=Ranges, nb_labels=0, line_width=0.7,
                           xlabel='cross range, m', ylabel='range, m',
                           zlabel='height, m')
            Ax.axes.font_factor = 1.0

    def PlotPhase(self):
        """
        James Elgy - 2019
        Function to plot the spatial frequency support of the complex SAR
        image.
        """
        try:
            self.ComplexImage
        except AttributeError:
            self.BackProjection_CPU()

        self.ImageSupport = np.fft.fftshift(np.fft.fftn(self.ComplexImage))
        PLog = 20*np.log10(np.abs(self.ImageSupport))
        PLog = np.squeeze(PLog - np.max(PLog))

        if len(PLog.shape) < 3:
            plt.rc('text', usetex=True)

            fig = plt.figure()
            plt.inferno()
            ax = fig.add_subplot(111)
            ax.imshow(PLog, vmin=np.max(PLog)-self.Params.DynamicRange,
                      vmax=np.max(PLog), origin='lower')
            ax.set_aspect('auto', 'box')
            ax.set_xlabel('$F_{x}$')
            ax.set_ylabel('$F_{y}$')
            ax.set_title('Image Support')
        else:

            VMax = np.max(PLog)
            VMin = VMax - self.Params.DynamicRange
            mlab.figure(bgcolor=(1, 1, 1), fgcolor=(0, 0, 0))
            mlab.pipeline.volume(mlab.pipeline.scalar_field(PLog),
                                 vmin=VMin, vmax=VMax)
            mlab.outline()
            Ax = mlab.axes(nb_labels=0, line_width=0.7,
                           xlabel='$F_{x}$', ylabel='$F_{y}$',
                           zlabel='$F_{z}$')
            Ax.axes.font_factor = 1.0

    def SaveVolumeImage(self, Dir):
        """
        James Elgy - 2019
        Function to save the complex sar data as a grayscale multipage tiff
        stack to the directory defined in Dir.
        The data is converted into db before saving.
        """
        Image = 20*np.log10(np.abs(self.ComplexImage))
        Image = Image - np.max(Image)
        Image[Image < -self.Params.DynamicRange] = -self.Params.DynamicRange
        NormImage = (Image - np.min(Image))/(np.max(Image) - np.min(Image))
        # NormImage = np.swapaxes(NormImage,0,2)
        skimage.external.tifffile.imsave(Dir, NormImage)

    
"""
Support Functions

"""


@jit([complex128(float64, float64, float64, float64, float64, float64, float64,
                 float64, complex128)], parallel=True)
def _BP_CPU(xm, ym, zm, RPvec, A1, A2, Freq, RangeSc, RangeProfile):

    # Preallocating Array
    ComplexImage = np.zeros(xm.shape, dtype=np.complex128)
    Freq = np.median(Freq)

    # For each position, calc dr and interpolate onto imaging grid
    for pos in prange(0, RangeProfile.shape[1]):

        # PercentDone = (100*pos)//RangeProfile.shape[1]
        # print(PercentDone, '%')

        dr_temp = np.sqrt((A1[0, pos] - xm)**2 +
                          (A1[1, pos] - ym)**2 +
                          (A1[2, pos] - zm)**2) \
                + np.sqrt((A2[0, pos] - xm)**2 +
                          (A2[1, pos] - ym)**2 +
                          (A2[2, pos] - zm)**2)
        dr = dr_temp/2 - RangeSc[pos]

        Correction = np.exp(-1j * 4 * Freq * dr * np.pi / c)

        for ind in range(0, dr.shape[0]):
            ComplexImage[ind, :] = ComplexImage[ind, :] + \
                                    np.conj(np.interp(dr[ind, :], RPvec,
                                                      RangeProfile[:, pos])) \
                                    * Correction[ind, :]

    return ComplexImage


def _BP_GPU(xm, ym, zm, RPvec, Ant1, Ant2, Freq, RangeSc, RangeProfileChunk):
	
	# This version of the backprojection is significantly slower than _BP_GPU_2.
	# However it is more stable. It is not called by default, however it has
	# been left in for completeness.
    OrigSize = xm.shape
    xm.shape = [np.prod(xm.shape), 1]
    ym.shape = [np.prod(ym.shape), 1]
    zm.shape = [np.prod(zm.shape), 1]

    ComplexImage = np.zeros(xm.shape, dtype=np.complex64)
    Pixel = np.ascontiguousarray(np.concatenate((xm, ym, zm), axis=1))
    Image = np.ascontiguousarray(np.zeros(xm.shape), dtype=np.complex64)

    TPB, BPG = _Initalise(np.prod(xm.shape))
    Pixel = cuda.to_device(Pixel)
    Image = cuda.to_device(Image)
    Freq = np.median(Freq)

    R0 = RPvec[0]
    Rdiff = np.diff(RPvec[0:2])

    for Pos in range(0, RangeProfileChunk.shape[1]):

        # PercentDone = (100*Pos)//RangeProfileChunk.shape[1]
        # print(PercentDone, '%')

        RP = np.conj(RangeProfileChunk[:, Pos])
        RP.shape = (RP.size, 1)
        RP = np.ascontiguousarray(RP)
        RangeSc_Pos = np.ascontiguousarray(RangeSc[Pos])

        A1 = Ant1[:, Pos]
        A2 = Ant2[:, Pos]
        A1.shape = (3, 1)
        A2.shape = (3, 1)
        A1 = np.ascontiguousarray(A1)
        A2 = np.ascontiguousarray(A2)

        _interpKernel[BPG, TPB](Pixel, A1, A2, RangeSc_Pos, Freq, R0,
                                Rdiff, RP, Image)

    ComplexImage = Image.copy_to_host()
    ComplexImage = np.reshape(ComplexImage, OrigSize)

    return ComplexImage


def _BP_GPU_2(xm, ym, zm, RPvec, Ant1, Ant2, Freq, RangeSc, RangeProfileChunk):

    OrigSize = xm.shape
    xm.shape = [np.prod(xm.shape), 1]
    ym.shape = [np.prod(ym.shape), 1]
    zm.shape = [np.prod(zm.shape), 1]

    ComplexImage = np.zeros(xm.shape, dtype=np.complex64)
    Pixel = np.ascontiguousarray(np.concatenate((xm, ym, zm), axis=1))
    Image = np.ascontiguousarray(np.zeros(xm.shape), dtype=np.complex64)

    TPB, BPG = _Initalise(np.prod(xm.shape))

    Pixel = cuda.to_device(Pixel)
    Image = cuda.to_device(Image)
    Freq = np.median(Freq)

    R0 = RPvec[0]
    Rdiff = np.diff(RPvec[0:2])

    # I've added in some functionality to split the range profiles up and dump
    # as much as possible onto the GPU. I've hardcoded a proportion cut off for
    # the free memory. I.e if the memory needed to dump the entire thing is
    # greater than 0.5* the amount of free memory, it will split the range
    # profiles up and iterate though.

    # Chunking Range Profiles
    Cont = cuda.current_context()
    Mem = Cont.get_memory_info()

    RPChunkSize = RangeProfileChunk.nbytes/(0.5*Mem[0])
    # Here, I'm using stream to set up an asynchronous execution.
    Stream = cuda.stream()
    if RPChunkSize <= 1:
        RP = cuda.to_device(np.ascontiguousarray(np.conj(RangeProfileChunk)), stream=Stream)
        RangeSc_Pos = cuda.to_device(np.ascontiguousarray(RangeSc), stream=Stream)
        A1 = Ant1
        A2 = Ant2
        A1 = cuda.to_device(np.ascontiguousarray(A1), stream=Stream)
        A2 = cuda.to_device(np.ascontiguousarray(A2), stream=Stream)

        _interpKernel2[BPG, TPB, Stream](Pixel, A1, A2, RangeSc_Pos, Freq, R0,
                                         Rdiff, RP, Image)
        ComplexImage = Image.copy_to_host()

    else:
        # This else statement is included so that datasets larger than the GPU
        # can be used.
        NChunks = np.array_split(RangeProfileChunk,
                                 int(math.ceil(RPChunkSize)), axis=1)

        for Index in range(0, int(math.ceil(RPChunkSize))-1):
            if Index == 0:
                CountOld = 0
                Count = NChunks[Index].shape[1]
            else:
                CountOld = Count
                Count = Count + NChunks[Index].shape[1]

            print(Count)
            RP_Contig = np.ascontiguousarray(np.conj(NChunks[Index]))
            RP = cuda.to_device(RP_Contig)
            RangeSc_Contig = np.ascontiguousarray(RangeSc[CountOld:Count])
            RangeSc_Pos = cuda.to_device(RangeSc_Contig)
            A1 = Ant1[:, CountOld:Count]
            A2 = Ant2[:, CountOld:Count]
            A1 = cuda.to_device(np.ascontiguousarray(A1))
            A2 = cuda.to_device(np.ascontiguousarray(A2))
            Stream.synchronize()
            _interpKernel2[BPG, TPB](Pixel, A1, A2, RangeSc_Pos, Freq,
                                     R0, Rdiff, RP, Image)

    Stream.synchronize()
    ComplexImage = np.reshape(ComplexImage, OrigSize)
    ComplexImage = np.squeeze(ComplexImage)

    return ComplexImage


def _Initalise(imsize):
    # Set the number of threads in a block
    threadsperblock = 32
    # Calculate the number of thread blocks in the grid
    blockspergrid = (imsize + (threadsperblock - 1)) // threadsperblock

    return (threadsperblock, blockspergrid)


@cuda.jit()
def _interpKernel2(Pixel, A1, A2, RangeSc, Freq, R0, Rdiff, RP, Image):
    # Second version of the backprojection kernel. This version has the for
    # loop moved inside the kernel. This allows for more data to be dumped onto
    # the gpu in one go, thus reducing transfer times.

    Ele = cuda.grid(1)
    PhaseConst = -1j * 4 * Freq * cmath.pi / c
    # if pos < Image.size:  # Check array boundaries
    for Pulse in range(0, RP.shape[1]):
        # Doing backprojection for each pixel in image.
        dr = (math.sqrt((A1[0, Pulse] - Pixel[Ele, 0])**2 +
                        (A1[1, Pulse] - Pixel[Ele, 1])**2 +
                        (A1[2, Pulse] - Pixel[Ele, 2])**2) +
              math.sqrt((A2[0, Pulse] - Pixel[Ele, 0])**2 +
                        (A2[1, Pulse] - Pixel[Ele, 1])**2 +
                        (A2[2, Pulse] - Pixel[Ele, 2])**2))/2 - RangeSc[Pulse]

        Correction = cmath.exp(PhaseConst * dr)

        # Linear interp section. Doing it like this avoids the search
        # algorithms used in numpy.
        Ind = int((dr - R0)//Rdiff[0])
        Rv_min = R0 + Ind*Rdiff[0]
        Rv_max = R0 + (Ind+1)*Rdiff[0]
        Grad = (RP[Ind+1, Pulse] - RP[Ind, Pulse]) / (Rv_max - Rv_min)
        Interpval = RP[Ind, Pulse] + (Grad*(dr - Rv_min))

        # Updating image.
        cuda.syncthreads()
        Image[Ele, 0] = Image[Ele, 0] + (Interpval * Correction)


@cuda.jit()
def _interpKernel(Pixel, A1, A2, RangeSc, Freq, R0, Rdiff, RP, Image):

    pos = cuda.grid(1)
    PhaseConst = -1j * 4 * Freq * cmath.pi / c
    if pos < Image.size:  # Check array boundaries
        # Doing backprojection for each pixel in image.
        dr = (math.sqrt((A1[0, 0] - Pixel[pos, 0])**2 +
                        (A1[1, 0] - Pixel[pos, 1])**2 +
                        (A1[2, 0] - Pixel[pos, 2])**2) +
              math.sqrt((A2[0, 0] - Pixel[pos, 0])**2 +
                        (A2[1, 0] - Pixel[pos, 1])**2 +
                        (A2[2, 0] - Pixel[pos, 2])**2))/2 - RangeSc[0]
        Correction = cmath.exp(PhaseConst * dr)

        # Linear interp section
        Ind = int((dr - R0)//Rdiff[0])
        Rv_min = R0 + Ind*Rdiff[0]
        Rv_max = R0 + (Ind+1)*Rdiff[0]
        Grad = (RP[Ind+1, 0] - RP[Ind, 0]) / (Rv_max - Rv_min)
        Interpval = RP[Ind, 0] + (Grad*(dr - Rv_min))

        # Updating image.
        Image[pos, 0] = Image[pos, 0] + (Interpval * Correction)
